import 'package:digital_3g_common/backend_switcher.dart';
import 'package:flutter/material.dart';
import 'package:flutter_bloc/flutter_bloc.dart';
import 'package:provider/provider.dart';
import 'package:uuid/uuid.dart';

import '../api/admin_api.dart';
import '../api/auth_api.dart';
import '../models/access_code.dart';
import 'content_dialog.dart';
import 'scope_selector.dart';

class CodeDialog extends StatefulWidget {
  const CodeDialog({Key? key}) : super(key: key);

  @override
  _CodeDialogState createState() => _CodeDialogState();
}

class _CodeDialogState extends State<CodeDialog> {
  AdminApi? _adminApi;

  int _amount = 1;
  String _prefix = '';

  Scope _scope = Scope.user;

  @override
  Widget build(BuildContext context) {
    return ContentDialog(
      title: const Text('Zugangscodes generieren'),
      children: [
        ListTile(
          title: const Text('Anzahl'),
          trailing: SizedBox(
            width: 48.0,
            child: TextField(
              controller: TextEditingController(text: _amount.toString()),
              keyboardType: TextInputType.number,
              onChanged: (String input) => _amount = int.tryParse(input) ?? 0,
            ),
          ),
        ),
        ListTile(
          title: const Text('Code-Prefix'),
          trailing: SizedBox(
            width: 156.0,
            child: TextField(
              controller: TextEditingController(text: _prefix),
              onChanged: (String input) => _prefix = input,
            ),
          ),
        ),
        ListTile(
          title: const Text('Rolle'),
          trailing: SizedBox(
            width: 128.0,
            child: ScopeSelector(onChanged: (Scope? scope) => _scope = scope!, value: _scope),
          ),
        ),
        const Divider(),
      ],
      actions: [
        ElevatedButton(
          onPressed: () => Navigator.of(context).pop<List<AccessCode>>(null),
          child: const Text('Abbrechen'),
        ),
        ElevatedButton(
          onPressed: _onSubmit,
          child: const Text('Erstellen'),
        ),
      ],
    );
  }

  @override
  void initState() {
    super.initState();
    _initApi();
  }

  Future<List<AccessCode>> onGenerateBatch() async {
    if (_amount == 0) return [];

    List<AccessCode> codes = [];
    Uuid uuid = const Uuid();
    for (int i = 0; i < _amount; i++) {
      codes.add(
        AccessCode(
            code: '${_prefix.isNotEmpty ? '$_prefix-' : ''}${uuid.v4()}',
            scope: _scope.toString().split('.').last),
      );
    }

    final result = await showDialog<bool>(
      context: context,
      builder: (BuildContext context) => ContentDialog(
        title: const Text('Zugangscodes wirklich erstellen?'),
        children: [
          Padding(
            padding: const EdgeInsets.all(8.0),
            child: Text(
                'Es werden folgende Codes für die Rolle "${_scope.toString().split('.').last}" eingefügt:'),
          ),
          Padding(
            padding: const EdgeInsets.all(16.0),
            child: TextField(
              maxLines: 7,
              controller: TextEditingController(
                text: codes.fold<String>(
                    '', (String previous, AccessCode i) => '$previous${i.code}\n'),
              ),
            ),
          ),
        ],
        actions: [
          ElevatedButton(
            onPressed: () => Navigator.pop(context, false),
            child: const Text('Nein'),
          ),
          ElevatedButton(
            onPressed: () => Navigator.pop(context, true),
            child: const Text('Ja'),
          ),
        ],
      ),
    );

    if (!(result ?? false)) return [];

    return await onSubmitBatch(codes);
  }

  Future<List<AccessCode>> onSubmitBatch(List<AccessCode> codes) async {
    final List<AccessCode> newCodes = [];
    for (var code in codes) {
      final response = await _adminApi!.addAccessCode(code.code, code.scope);
      if (response != null) {
        newCodes.add(response);
      }
    }
    ScaffoldMessenger.of(context)
        .showSnackBar(SnackBar(content: Text('${codes.length} Zugangscodes wurden eingefügt')));
    return newCodes;
  }

  void _initApi() async {
    final _client = Provider.of<AuthApi>(context, listen: false).client;
    _adminApi = AdminApi(_client, BlocProvider.of<BackendSwitcher>(context).state);
    super.didChangeDependencies();
  }

  void _onSubmit() async {
    final batch = await onGenerateBatch();

    if (batch.isEmpty) return;

    Navigator.of(context).pop<List<AccessCode>>(batch);
  }
}
